/*
  By accepting this notice, you agree to be bound by the following
  agreements:
  
  ufdbGuard is copyrighted (C) 2005,2006 by URLfilterDB with all rights reserved.
  ufdbGuard is based on squidGuard.

  squidGuard is copyrighted (C) 1998 by
  ElTele st AS, Oslo, Norway, with all rights reserved.
  
  This program is free software; you can redistribute it and/or modify it
  under the terms of the GNU General Public License (version 2) as
  published by the Free Software Foundation.  It is distributed in the
  hope that it will be useful, but WITHOUT ANY WARRANTY; without even the
  implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
  PURPOSE.  See the GNU General Public License (GPL) for more details.
  
  You should have received a copy of the GNU General Public License
  (GPL) along with this program.
*/


#include "sg.h"
#include <signal.h>
#include <pthread.h>

#include "ufdblib.h"

extern int sig_hup;           
extern int sig_other;           
extern int sig_alrm;           
extern int globalDebugTimeDelta;
extern char ** globalArgv;           
extern char ** globalEnvp;           
extern struct Acl * defaultAcl;
extern struct LogFileStat * LogFileStat;

extern struct Source * Source;
extern struct Destination * Dest;



void sgSetSignalHandler( int signal, void (*handler)(int) )
{
#if HAVE_SIGACTION
  struct sigaction act;

#ifndef SA_NODEFER
#define SA_NODEFER 0
#endif
  act.sa_handler = handler;
  act.sa_flags = SA_NODEFER | SA_RESTART;
  sigemptyset( &act.sa_mask );
  sigaction( signal, &act, NULL );

#else

#if HAVE_SIGNAL
  signal( signal, handler );
#else
  ufdbLogError( "sgSetSignalHandler: cannot set signals" );
#endif

#endif
}

void sgHandlerSigHUP( int s )
{
  if (sig_hup == s)
     abort();
  sig_hup = s;
}

void sgSignalHandler( int s )
{
  if (sig_other == s)
     abort();
  sig_other = s;
}

void sgReloadConfig( void )
{
  struct LogFileStat * sg;

  sig_hup = 0;
  ufdbLogError( "got sigHUP reload config" );

  /* closing logfiles */
  for (sg = LogFileStat; sg != NULL; sg = sg->next)
  {
    if (sg->fd == stderr || sg->fd == stdout)
      continue;
    fclose( sg->fd );
  }

  execve( *globalArgv, globalArgv, globalEnvp );
  fprintf( stderr, "error execve: %d\n", errno );
}


void sgAlarm( int signal )
{
  sig_alrm = 1;
  sgTimeNextEvent();
}


#if 0	
static void printSquidInfo( struct SquidInfo * s )
{
  fprintf( stderr, "   protocol = %s\n", s->protocol );
  fprintf( stderr, "   port     = %d\n", s->port     );
  fprintf( stderr, "   dot      = %d\n", s->dot      );
  UFDBprintRevURL( s->revUrl );
  fprintf( stderr, "   domain   = %s\n", s->domain       );
  fprintf( stderr, "   orgdomain= %s\n", s->orig_domain  );
  fprintf( stderr, "   url      = %s\n", s->url          );
  fprintf( stderr, "   orig     = %s\n", s->orig         );
  fprintf( stderr, "   surl     = %s\n", s->surl         );
  fprintf( stderr, "   src      = %s\n", s->src          );
  fprintf( stderr, "   srcDom   = %s\n", s->srcDomain    );
  fprintf( stderr, "   ident    = %s\n", s->ident        );
  fprintf( stderr, "   method   = %s\n", s->method       );
}
#endif


/*
 * parsers the squidline:
 * URL ip-address/fqdn ident method
 * http://www.sex.com 10.1.1.44/- - GET
 */
int parseLine( 
  UFDBthreadAdmin *  admin, 
  char *             line, 
  struct SquidInfo * s )
{
  char * p, *d, *a, *e, *o, *field;
  char * up, *upp;
  int    i; 
  char * lineptr;

  s->revUrl = NULL;

#if 0
  fprintf( stderr, "   line: %s\n", line );
#endif

  field = strtok_r( line, "\t ", &lineptr );
  if (field == NULL)
    return 0;

  s->quota = NULL;

  strcpy( s->orig, field );
  for (p = field; *p != '\0'; p++) 	/* convert url to lowercase chars */
  {
    if (*p >= 'A'  &&  *p <= 'Z')
      *p += 'a' - 'A';
  }

  s->url[0] = s->domain[0] = s->orig_domain[0] = s->src[0] = s->ident[0] = s->ident[1] =
    s->method[0] = s->srcDomain[0] = s->surl[0] =  '\0';
  s->dot = 0;
  s->port = 80;

  p = strstr(field,"://");
  if (p == NULL)		/* no protocol, defaults to http */
  {
    strcpy( s->protocol, "http" );
    p = field;
  }
  else
  {
    *p = '\0';			   	/* uhoh, I can not write into the URL string !! */
    i = p - field;
    if (i > 15)
       i = 15;				/* WHOEHA a large protocol name. truncate it */
    strncpy( s->protocol, field, i );
    s->protocol[i] = '\0';
    if (strcasecmp( s->protocol, "https" ) == 0)
       s->port = 443;
    *p = ':';				/* fix the \0 */
    p += 3;
    field = p;
  }

  /* do some url decoding */

  up = field;
  upp = s->url;
  i = 0;
  while(up[i] != '\0') 
  {
    if (up[i] == '%')
    {
      if (isxdigit( (int) up[i+1]) && isxdigit( (int) up[i+2]))
      {
	int    hex;
	char   h1, h2;
	/* note that the URL is already converted to lowercase */
#if 0
        h1 = up[i+1] >= 'a' ? up[i+1] - ('a' - 'A') : up[i+1];
        h2 = up[i+2] >= 'a' ? up[i+2] - ('a' - 'A') : up[i+2];
        hex = h1 >= 'A' ? h1 - 'A' - 10 : h1 - '0';
        hex *= 16;
        hex += h2 >= 'A' ? h2 - 'A' - 10 : h2 - '0';
#else
	h1 = up[i+1];
	h2 = up[i+2];
	hex  = (h1 <= '9') ? h1 - '0' : h1 - 'a' + 10;
        hex *= 16;
	hex += (h2 <= '9') ? h2 - '0' : h2 - 'a' + 10;
#endif
	/* don't convert control characters */
	if (hex <= 0x20)
	{
	   *upp++ = up[i++];
	   *upp++ = up[i++];
	   *upp++ = up[i];
	}
	else
	{ 
	   if (hex >= 'A'  &&  hex <= 'Z')
	      hex += 'a' - 'A';
	   *upp++ = hex;
	   i += 2;
	}
      }
      else
      { /* an erroneous hex code, we ignore it */
        *upp++ = up[i++];
        *upp++ = up[i++];
        *upp++ = up[i];
      }
    }
    else
    {
      while (up[i] == '/'  &&  up[i+1] == '/')		/* substitute // by / */
         i++;
      *upp++ = up[i];
    }
    i++;
  }
  *upp++ = up[i];
  *upp = '\0';

#if 0
  fprintf( stderr, ">>>> s->url = %s\n", s->url );
#endif

  d = strchr(p,'/'); 		/* find domain end */
  e = d;
  a = strchr(p,'@'); 		/* find auth  */
  if (a != NULL  &&  (a < d || d == NULL)) 
    p = a + 1;

  a = strchr(p,':'); 		/* find port */
  if (a != NULL  &&  (a < d || d == NULL))
  {
    s->port = 0;
    o = a + 1;
    while (isdigit( (int) *o))
    {
       s->port = s->port * 10 + (*o - '0');
       o++;
    }
    strcpy( a, o );		/* remove the port number from the URL */
    e = a;
  }

  if (e == NULL)
     strcpy( s->orig_domain, p );
  else
  {
     i = e - p;
     strncpy( s->orig_domain, p, i );
     s->orig_domain[i] = '\0';
  }
  
  /* strip prefix www. www8. ftp. ftp5. ... */
  o = p;

  if ((p[0] == 'w' && p[1] == 'w' && p[2] == 'w') ||		/* match www. */
      (p[0] == 'f' && p[1] == 't' && p[2] == 'p'))  		/* match ftp. */
  {
    p += 3;
    while (p[0] <= '9' && p[0] >= '0')              		/* match www99. */
      p++;
    if (p[0] != '.'  ||  strchr( p+1, '.' ) == NULL)		/* skip skip www_xxx.com and www99.com */
      p = o;  /* not a hostname */
    else
      p++;
  }

  if (e == NULL)
  {
    strcpy( s->domain, p );		/* UFDB removes www. prefix from domains, squidGuard not */
  }
  else 
  {
    memcpy( s->domain, p, e - p );	/* UFDB removes www. prefix from domains, squidGuard not */
    *(s->domain + (e - p)) = '\0';
  }
  strcpy( s->surl, p );

#if 0
  fprintf( stderr, ">>>>> s->surl = %s\n", s->surl );
#endif

  for (o = s->domain; isdigit( (int) *o) || *o == '.'; o++)
     ;
  if (*o == '\0')
    s->dot = 1;

  if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)
  {
    o = strchr(p,'/');
    if (o != NULL) 
    {
      memcpy( s->src, p, o-p );
      s->src[o-p] = '\0';
      strcpy( s->srcDomain, o+1 );
      if (*s->srcDomain == '-')
	s->srcDomain[0] = '\0';
    } else
      strcpy( s->src, p );

    if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)
    {
      if (p[0] != '-'  ||  p[1] != '\0')
      {
	strcpy( s->ident, p );
	for (p = s->ident; *p != '\0'; p++) 	/* convert ident to lowercase chars */
	{
	  if (*p >= 'A'  &&  *p <= 'Z')
	    *p += 'a' - 'A';
        }
      }

      if ((p = strtok_r(NULL," \t\n",&lineptr)) != NULL)
         strcpy( s->method, p );
    }
  }

  if (s->domain[0] == '\0')
    return 0;
  if (s->method[0] == '\0')
    return 0;

#if 0
  fprintf( stderr, ">>>>>> s->surl = %s\n", s->surl );
#endif

  /* s->revUrl is pointing to malloced memory and must be freed at the end of the main loop ! */
  s->revUrl = UFDBgenRevURL( admin, (unsigned char *) s->surl );

#if 0
  printSquidInfo( s );
#endif

  return 1;
}


/* 
 *  returns a pointer to the domain part of a fully-qualified  hostname
 *  so www.abc.xyz.dom/index.html -> xyz.dom/index.html
 */
char * sgSkipHostPart( char * domain )
{
  char *p = domain , *d1 = NULL, *d2 = NULL, *path = NULL;

  if ((path = (char *) strchr(p,'/')) == NULL)
    path = domain; 

  while((p = (char *) strchr(p,'.')) != NULL )
  {
    if (p > path && path != domain)
      break;
    d2 = d1;
    d1 = p;
    p++;
  }
  if (d2 != NULL)
    return d2+1;

  return domain;
}

/*
 * checks the vality of a dotted address. 
 */
ulong * sgConvDot( char * dot, ulong * ipaddr )
{
   char * t;
   int    octet;
   int    shift;

   *ipaddr = 0;
   shift = 24;
   t = dot;
   while (*t != '\0')
   {
      if (*t > '9'  ||  *t < '0')
         return NULL;

      octet = 0;
      while (*t >= '0'  &&  *t <= '9') 
      {
	 octet = octet*10 + (*t - '0');
         t++;
      }

      if (*t == '.') 
         t++;
      else if (*t != '\0') 
         return NULL;

      if (shift < 0)
         return NULL;

      if (octet > 255)
         return NULL;

      *ipaddr |= octet << shift;
      shift -= 8;
   }

   return ipaddr;
}


/*
 * Reverses cmp of strings
 */
int sgStrRcmp( char * a, char * b )
{
  char *a1 = (char *) strchr(a,'\0');
  char *b1 = (char *) strchr(b,'\0');

  while(*a1 == *b1)
  {
    if (b1 == b || a1 == a)
      break;
    a1--; b1--;
  }
  if (a1 == a && b1 == b)
    return *a1 - *b1; 
  if (a1 == a)
    return -1;
  if (b1 == b)
    return 1;

  return *a1 - *b1;
}


int sgStrRncmp( char * a, char * b, int blen )
{
  char *a1 = (char *) strchr(a,'\0');
  char *b1 = (char *) strchr(b,'\0');

  while(*a1 == *b1 && blen > 0 )
  {
    if (b1 == b || a1 == a)
      break;
    a1--; b1--; blen--;
  }

  if (a1 == a && b1 == b)
    return *a1 - *b1; 

  if (blen == 0)
    return *a1 - *b1;

  if (a1 == a)
    return -1;

  if (b1 == b)
    return 1;

  return *a1 - *b1;
}


/*
 *  sgDomStrRncmp checks if p2 is equal to or a subdomain of p1
 */
int sgDomStrRcmp( char * p1, char * p2 )
{
  char *p1e = (char *) strchr(p1,'\0');
  char *p2e = (char *) strchr(p2,'\0');

  for(;p1e >= p1 && p2e >= p2 && *p1e == *p2e; p1e--, p2e--)
     ;

  if (p1e < p1 && p2e < p2)
    return 0;
  if (p2e < p2)
    return -*p1e;
  if (p1e < p1 && *p2e == '.')	/* found a subdomain */
    return 0;
  return *p1e - *p2e;
}


/*
 *  Regexp functions
 */


char * ufdbRegExpSubst( struct ufdbRegExp * regexp, char * pattern, char * newstring )
{
  struct ufdbRegExp * re;
  regmatch_t        pm;
  char *            result = NULL;
  char *            p;
  int               substlen;

  newstring[0] = '\0';
  for (re = regexp; re != NULL; re = re->next)
  {
    if (regexec(re->compiled, pattern, 1, &pm, 0) != 0) 
    {
      result = NULL;
    }
    else
    {
      substlen = strlen( re->substitute );
      if (re->httpcode != NULL)
	strcpy( newstring, re->httpcode );
      else
	*newstring = '\0';
      p = newstring;
      do {
	if ((p - newstring) + pm.rm_so >= MAX_BUF)
	  break;
	p = strncat( newstring, pattern, pm.rm_so );
	if ((p - newstring) + substlen >= MAX_BUF)
	  break;
	p = strcat( newstring, re->substitute );	
	pattern = pattern + pm.rm_eo;
      } while (regexec(re->compiled, pattern, 1, &pm, REG_NOTBOL) == 0 && re->global);
      if ((p - newstring) + strlen(pattern) <= MAX_BUF)
	p = strcat( newstring, pattern );
      result = newstring;
      break;
    }
  }

  return result;
}


char * sgRewriteExpression( struct sgRewrite * rewrite, char * subst, char * newstring )
{
  char * result;

  result = ufdbRegExpSubst( rewrite->rewrite, subst, newstring );
  return result;
}


/*
 *  Expand a redirection URL string: fill in the %i, %n etc.
 */
char * sgParseRedirect( 
  char *             redirect,
  struct SquidInfo * req,
  struct Acl *       acl, 
  struct AclDest *   aclpass,
  char *             buf )
{
  char * p;
  char * q;
  char * t;
  char * d;

#ifdef UFDB_USERQUOTA_SUPPORT
  struct Source * s = req->lastActiveSource;
#endif

  buf[0] = '\0';
  if (aclpass == NULL)
    aclpass = defaultAcl->pass;

  q = NULL;
  p = redirect;
  while((p = strchr(p,'%')) != NULL)
  {
    if (q == NULL) {
      memcpy( buf, redirect, p - redirect );
      buf[p - redirect] = '\0';
    } else {
      strncat( buf, q,  p - q );
    }

    p++;
    switch (*p)
    {
    case 'a': 			/* Source Address */
      strcat( buf, req->src );
      break;

    case 'A':
       t = sgSettingGetValue( "administrator" );
       if (t == NULL)
          t = "your system administrator or helpdesk";
       d = buf + strlen( buf );
       while (*t != '\0')
       {
          if (*t == ' ' || *t == '=' || *t == '&' || *t == '\\' || *t == '"')
	  {
	     char h;

	     *d++ = '%';
	     h = *t / 16;
	     *d++ = (h >= 10) ? ('A' - 10 + h) : ('0' + h);
	     h = *t % 16;
	     *d++ = (h >= 10) ? ('A' - 10 + h) : ('0' + h);
	  }
	  else
	     *d++ = *t;
	  t++;
       }
       *d = '\0';
       break;

    case 'i': 			/* Source User Ident */
      if ((req->ident[0] == '\0') || (req->ident[0] == '-'  &&  req->ident[1] == '\0'))
      {
	strcat( buf, "unknown" );
      }
      else
      {
	strcat( buf, req->ident );
      }
      break;

    case 'q': 				/* userquota info */
#ifdef UFDB_USERQUOTA_SUPPORT
      if (s != NULL  && 
          s->userquota.seconds != 0  &&  
	  strcmp(req->ident, "-") != 0) 
      {
	struct UserQuotaInfo * userquota;
	if (sgDbLookup( s->userDb, req->ident, (char **) (char *) &userquota ) == 1) 
	{
	  char qbuf[150];
	  sprintf( qbuf, "%d-%d-%d-%d-%d-%d", 
		   s->userquota.renew, (int) s->userquota.seconds, userquota->status,
		   (int) userquota->time, (int) userquota->last, userquota->consumed );
	  strcat(buf, qbuf);
	}
	else
	{
	  strcat( buf, "noquota" );
	}
      }
      else
      {
	strcat( buf, "noquota" );
      }
#else
      strcat( buf, "noquota" );
#endif  /* UFDB_USERQUOTA_SUPPORT */
      break;

    case 'n': 		/* Source Domain Name */
      if ((req->srcDomain[0] == '\0') || (req->srcDomain[0] == '-'  &&  req->srcDomain[1] == '\0'))
      {
	strcat(buf, "unknown" );
      }
      else
      {
	strcat( buf, req->srcDomain );
      }
      break;

    case 'p': 		/* The url path */
      if ((t = strstr(req->orig,"//")) != NULL)
      {
        t += 2;
        if ((t = strchr(t,'/')) != NULL)
	{
	  t++;
	  strcat( buf, t );
        }
      }
      break;

    case 'f': 		/* The url file */
      if ((t = strrchr(req->orig,'/')) != NULL)
      {
        t++;
        strcat( buf, t );
      }
      break;

    case 's': 		/* Source Class Matched */
      if (acl->source == NULL || acl->source->name == NULL)
	strcat( buf, "default" );
      else
	strcat( buf, acl->source->name );
      break;

    case 't': 		/* Target Class Matched */
      if (aclpass == NULL)
	strcat( buf, "unknown" );
      else if (aclpass->name == NULL)
	if (aclpass->type == ACL_TYPE_INADDR)
	  strcat( buf, "in-addr" );
	else if (aclpass->type == ACL_TYPE_TERMINATOR)
	  strcat( buf, "none" );
	else
	  strcat( buf, "unknown" );
      else
	strcat( buf, aclpass->name );
      break;

    case 'u': 		/* Requested URL */
      strcat( buf, req->orig );
      break;

    default:
      strcat( buf, "%" );

    }
    p++;
    q = p;
  }

  if (buf[0] == '\0')
    q = redirect;
  else 
    q = buf;

  return q;
}


void sgEmergency( void )
{
  char buf[MAX_BUF];

  ufdbLogError("going into emergency mode");
  while(fgets(buf, MAX_BUF, stdin) != NULL)
  {
    puts("");
    fflush(stdout);
  }
  ufdbLogError("ending emergency mode, stdin empty");
  exit(1);
}


/*
 *  converts yyyy.mm.ddTHH:MM:SS to seconds since EPOC
 */

time_t iso2sec( char * date )
{
  struct tm *t;
  int y,m,d,H,M,S;

  t = (struct tm *) ufdbCalloc(1,sizeof(struct tm));
  sscanf(date,"%4d%*[.-]%2d%*[.-]%2d%*[T]%2d%*[:-]%2d%*[:-]%2d",
	 &y,&m,&d,&H,&M,&S);
  m--; 
  y = y - 1900;
  if (y < 0 || m < 0 || m > 11 || d < 1 || d > 31 || H < 0 || H > 23 
     || M < 0 || M > 59 || S < 0 || S > 59)
    return (time_t) -1;
  t->tm_year = y;
  t->tm_mon =  m;
  t->tm_mday = d;
  t->tm_hour = H;
  t->tm_min = M;
  t->tm_sec = S;

  return (time_t) mktime(t);
}

/*
 *  converts yyyy.mm.dd to seconds since EPOC
 */
time_t date2sec( char * date )
{
  struct tm *t;
  int y,m,d;

  t = (struct tm *) ufdbCalloc(1,sizeof(struct tm));
  sscanf(date,"%4d%*[.-]%2d%*[.-]%2d",&y,&m,&d);
  m--; 
  y = y - 1900;
  if (y < 0 || m < 0 || m > 11 || d < 1 || d > 31)
    return (time_t) -1;
  t->tm_year = y;
  t->tm_mon =  m;
  t->tm_mday = d;

  return (time_t) mktime(t);
}


void niso( time_t t, char * buf )
{
  time_t    tp;
  struct tm lc;

  if (t == 0)
    tp = time(NULL) + globalDebugTimeDelta;
  else
    tp = t;
  localtime_r( &tp, &lc );
  sprintf( buf, "%04d-%02d-%02d %02d:%02d:%02d", lc.tm_year + 1900, lc.tm_mon + 1,
	  lc.tm_mday, lc.tm_hour, lc.tm_min, lc.tm_sec );
}


struct UserQuotaInfo * setuserquota( struct UserQuotaInfo * uq )
{
  uq->time = 0; 
  uq->last = 0; 
  uq->consumed = 0; 
  uq->status = 0; 

  return uq;
}

